import jax
from functools import partial

from jf1uids.geometry.boundaries import _boundary_handler
from jf1uids.physics_modules.stellar_wind.stellar_wind import wind_source


@partial(jax.jit, static_argnames=['config'])
def add_physical_sources(primitive_state, dt, config, params, helper_data):
    
    # stellar wind
    if config.stellar_wind:
        source_term = wind_source(params.wind_params, primitive_state, helper_data.volumetric_centers, config.dx)
        primitive_state = primitive_state.at[:, :].add(dt * source_term)

    primitive_state = _boundary_handler(primitive_state, config.left_boundary, config.right_boundary)

    return primitive_state



from typing import NamedTuple
import jax.numpy as jnp

class CheckpointData(NamedTuple):
    # Checkpoint data
    times: jnp.ndarray = None
    total_mass_proxy: jnp.ndarray = None
    total_energy_proxy: jnp.ndarray = None
    states: jnp.ndarray = None
    current_checkpoint: int = 0
    num_iterations: int = 0

    runtime: float = 0.0


from typing import NamedTuple

from jf1uids.physics_modules.stellar_wind.stellar_wind import WindParams

# Different from the simulation configuration, the simulation parameters
# do not require recompilation when changed.

# Examples are the CFL number or settings of any physics module
# that might be implemented.

class SimulationParams(NamedTuple):
    C_cfl: float = 0.8
    gamma: float = 5/3

    dt_max: float = 0.001
    t_end: float = 0.2

    # parameters of physics modules
    wind_params: WindParams = WindParams()


from jf1uids.data_classes.simulation_checkpoint_data import CheckpointData
from jf1uids.data_classes.simulation_helper_data import HelperData
from jf1uids.data_classes.simulation_helper_data import get_helper_data


import numpy as np
from astropy.constants import M_sun
from astropy import units as u
from scipy.integrate import solve_ivp

# class for the Weaver et al. (1977) stellar wind solution
# initiated with the
#          terminal velocity v_inf
#          mass loss rate M_dot
#          background density rho_0
# which can be used to calculate the density and velocity at time t.
class Weaver:

    def __init__(self, v_inf, M_dot, rho_0, p_0, num_xi = 100, gamma = 5/3):

        # input wind parameters
        ## terminal velocity of the wind
        self.v_inf = v_inf
        ## mass loss rate of the wind
        self.M_dot = M_dot
        ## background density of the ISM
        self.rho_0 = rho_0
        ## background pressure of the ISM
        self.p_0 = p_0

        # derived wind parameters
        ## mechanical luminosity of the wind
        self.L_w = 0.5 * M_dot * v_inf ** 2

        # constants
        self.xi_crit = 0.86
        self.alpha = 0.88
        self.gamma = gamma

        # number of points to calculate the shell profiles
        self.num_xi = num_xi

        # calculate the shell profiles
        self.calculate_shell_profiles()

    def calculate_shell_profiles(self):
        # integrate equations 6 to 7 in weaver II, with
        # 3 * (U - xi) * U' - 2U + 3 P' / G = 0
        # (U - xi) * G' / G + U' + 2U / xi = 0
        # 3 * (U - xi) * P' - 3 * gamma * P * (U - xi) * G' / G - 4P = 0
        # where ' denotes \partial_\xi
        # from xi = 1 to xi = xi_crit with G(1) = 4, U(1) = 3/4, P(1) = 3/4
        # r = R_2 * xi; rho = rho_0 * G(xi)
        # v = V_2 * U(xi); p = rho_0 * V_2^2 * P(xi)

        gamma = self.gamma
        
        # the equations are given by
        def shell_equation(xi, y):
            G, U, P = y
            G_prime = 2*(-2*xi**2*G*U + 5*xi*G*U**2 + 2*xi*P - 3*G*U**3)*G/(3*xi*(gamma*xi*P - gamma*P*U - xi**3*G + 3*xi**2*G*U - 3*xi*G*U**2 + G*U**3))
            U_prime = 2*(-3*gamma*P*U + xi**2*G*U - xi*G*U**2 + 2*xi*P)/(3*xi*(gamma*P - xi**2*G + 2*xi*G*U - G*U**2))
            P_prime = 2*(-2*gamma*xi*U + 3*gamma*U**2 + 2*xi**2 - 2*xi*U)*G*P/(3*xi*(gamma*P - xi**2*G + 2*xi*G*U - G*U**2))
            return [G_prime, U_prime, P_prime]
        
        # initial conditions
        G1 = 4
        U1 = 3/4
        P1 = 3/4

        sol = solve_ivp(shell_equation, [1, self.xi_crit], [G1, U1, P1], t_eval=np.linspace(1, self.xi_crit, self.num_xi))

        self.xi = np.flip(sol.t)
        self.U = np.flip(sol.y[1])
        self.G = np.flip(sol.y[0])
        self.P = np.flip(sol.y[2])

    # returns the inner shock radius R_1
    def get_inner_shock_radius(self, t):
        return 0.9 * self.alpha ** 1.5 * (1 / self.rho_0 * self.M_dot) ** (3/10) * self.v_inf ** (1/10) * t ** (2/5)
    
    # returns the outer shock radius R_2
    def get_outer_shock_radius(self, t):
        return self.alpha * (self.L_w * t ** 3 / self.rho_0) ** (1/5)
    
    # returns the critical radius R_c
    def get_critical_radius(self, t):
        return self.xi_crit * self.get_outer_shock_radius(t)
    
    def get_radial_range_wind_interior(self, delta_R, t):
        R_1 = self.get_inner_shock_radius(t)
        R_c = self.get_critical_radius(t)
        return np.arange(R_1.to(u.parsec).value, R_c.to(u.parsec).value, delta_R.to(u.parsec).value) * u.parsec
    
    def get_radial_range_free_wind(self, delta_R, t):
        R_1 = self.get_inner_shock_radius(t)
        return np.arange(delta_R.to(u.parsec).value, R_1.to(u.parsec).value, delta_R.to(u.parsec).value) * u.parsec
    
    def get_radial_range_undisturbed_ism(self, delta_R, R_max, t):
        R_2 = self.get_outer_shock_radius(t)
        return np.arange(R_2.to(u.parsec).value, R_max.to(u.parsec).value, delta_R.to(u.parsec).value) * u.parsec
    
    # get inner pressure
    def get_pressure_profile(self, delta_R, R_max, t):

        # free wind profile (?????)
        # Rs_free_wind = self.get_radial_range_free_wind(delta_R, t)
        # pressures_free_wind = 1/20 * self.M_dot * self.v_inf / (4 * np.pi * Rs_free_wind ** 2)

        # wind interior
        Rs_wind_interior = np.array([self.get_inner_shock_radius(t).to(u.parsec).value, self.get_critical_radius(t).to(u.parsec).value]) * u.parsec
        pressure_wind_interior = 5 / (22 * np.pi * (self.xi_crit * self.alpha) ** 3) * (self.L_w ** 2 * self.rho_0 ** 3) ** (1/5) * t ** (-4/5) * np.array([1, 1])
        
        # shell
        # as of the RH-jump conditions at a contact discontinuity,
        # the pressure is continuous across the contact discontinuity
        Rs_shell = self.xi * self.get_outer_shock_radius(t)
        V2 = 15 / 25 * self.get_critical_radius(t) / t
        pressure_shell = self.rho_0 * V2 ** 2 * self.P
        pressure_shell = pressure_shell / pressure_shell[0] * pressure_wind_interior[1] # why is this necessary? sth wrong with the solution??

        # undisturbed ISM
        Rs_undisturbed_ism = self.get_radial_range_undisturbed_ism(delta_R, R_max, t)
        pressure_undisturbed_ism = self.p_0 * np.ones(len(Rs_undisturbed_ism))

        return np.concatenate((Rs_wind_interior, Rs_shell, Rs_undisturbed_ism)), np.concatenate((pressure_wind_interior, pressure_shell, pressure_undisturbed_ism))


    # get velocity profile
    def get_velocity_profile(self, delta_R, R_max, t):

        # free wind profile
        Rs_free_wind = self.get_radial_range_free_wind(delta_R, t)
        velocities_free_wind = self.v_inf * np.ones(len(Rs_free_wind))

        # wind interior
        Rs_wind_interior = self.get_radial_range_wind_interior(delta_R, t)
        R_c = self.get_critical_radius(t)
        velocities_wind_interior = 11 / 25 * R_c ** 3 / (Rs_wind_interior ** 2 * t) + 4 / 25 * Rs_wind_interior / t

        # shell
        Rs_shell = self.xi * self.get_outer_shock_radius(t)
        # by the RH-jump conditions at a contact discontinuity,
        # the velocity at the beginning of the shell is the 
        # velocity at the end of the wind interior, so
        V2 = 15 / 25 * R_c / t
        velocities_shell = V2 * self.U / self.U[0]

        # undisturbed ISM
        # here the velocity is 0
        Rs_undisturbed_ism = self.get_radial_range_undisturbed_ism(delta_R, R_max, t)
        velocities_unisturbed_ism = np.zeros(len(Rs_undisturbed_ism))

        return np.concatenate((Rs_free_wind, Rs_wind_interior, Rs_shell, Rs_undisturbed_ism)), np.concatenate((velocities_free_wind, velocities_wind_interior, velocities_shell, velocities_unisturbed_ism))
        
    # get density profile
    def get_density_profile(self, delta_R, R_max, t):
        # free wind profile
        Rs_free_wind = self.get_radial_range_free_wind(delta_R, t)
        densities_free_wind = self.M_dot / (4 * np.pi * Rs_free_wind ** 2 * self.v_inf)

        # profile in the wind interior
        Rs_wind_interior = self.get_radial_range_wind_interior(delta_R, t)
        R_c = self.get_critical_radius(t)
        densities_wind_interior = 0.628 * (self.M_dot ** 2 * self.rho_0 ** 3 * self.v_inf ** (-6)) ** (1 / 5) * t ** (-4 / 5) * (1 - (Rs_wind_interior / R_c) ** 3) ** (-8 / 33)

        # profile in the shell
        Rs_shell = self.xi * self.get_outer_shock_radius(t)
        densities_shell = self.rho_0 * self.G

        # profile in the undisturbed ISM
        Rs_undisturbed_ism = self.get_radial_range_undisturbed_ism(delta_R, R_max, t)
        densities_unisturbed_ism = self.rho_0 * np.ones(len(Rs_undisturbed_ism))

        return np.concatenate((Rs_free_wind, Rs_wind_interior, Rs_shell, Rs_undisturbed_ism)), np.concatenate((densities_free_wind, densities_wind_interior, densities_shell, densities_unisturbed_ism))



# constants, TODO: bundle
from jf1uids.geometry.boundaries import OPEN_BOUNDARY, REFLECTIVE_BOUNDARY
from jf1uids.geometry.geometry import SPHERICAL

# setup
from jf1uids.option_classes.simulation_config import SimulationConfig
from jf1uids.data_classes.simulation_helper_data import get_helper_data
from jf1uids.option_classes.simulation_params import SimulationParams

# module-setup
from jf1uids.physics_modules.stellar_wind.stellar_wind import WindParams

# run
from jf1uids.time_stepping.time_integration import time_integration

# postprocessing
from jf1uids.shock_finder.shock_finder import strongest_shock_radius
from jf1uids.fluid_equations.fluid import primitive_state

# units
from jf1uids.units import CodeUnits


import jax.numpy as jnp
import jax
from jf1uids.fluid_equations.euler import _euler_flux
from jf1uids.fluid_equations.fluid import conserved_state, speed_of_sound

@jax.jit
def _hll_solver(primitives_left, primitives_right, gamma):
    """
    Returns the conservative fluxes.
    """
    rho_L, u_L, p_L = primitives_left
    rho_R, u_R, p_R = primitives_right

    # calculate the sound speeds
    c_L = speed_of_sound(rho_L, p_L, gamma)
    c_R = speed_of_sound(rho_R, p_R, gamma)

    # get the left and right states and fluxes
    fluxes_left = _euler_flux(primitives_left, gamma)
    fluxes_right = _euler_flux(primitives_right, gamma)
    
    # very simple approach for the wave velocities
    wave_speeds_right_plus = jnp.maximum(jnp.maximum(u_L + c_L, u_R + c_R), 0)
    wave_speeds_left_minus = jnp.minimum(jnp.minimum(u_L - c_L, u_R - c_R), 0)

    # get the left and right conserved variables
    conservatives_left = conserved_state(primitives_left, gamma)
    conservatives_right = conserved_state(primitives_right, gamma)

    # calculate the interface HLL fluxes
    # F = (S_R * F_L - S_L * F_R + S_L * S_R * (U_R - U_L)) / (S_R - S_L)
    fluxes = (wave_speeds_right_plus * fluxes_left - wave_speeds_left_minus * fluxes_right + wave_speeds_left_minus * wave_speeds_right_plus * (conservatives_right - conservatives_left)) / (wave_speeds_right_plus - wave_speeds_left_minus)

    return fluxes


from typing import NamedTuple
import jax.numpy as jnp
import jax
from jf1uids.fluid_equations.fluid import pressure_from_energy

class WindParams(NamedTuple):
    wind_mass_loss_rate: float = 0.0
    wind_final_velocity: float = 0.0

@jax.jit
def wind_source(wind_params, primitive_state, r, dr):

    source_term = jnp.zeros_like(primitive_state)

    num_injection_cells = 10

    r_inj = r[num_injection_cells + 2]
    V = 4/3 * jnp.pi * r_inj**3

    drho_dt = wind_params.wind_mass_loss_rate / V
    dE_dt = 0.5 * wind_params.wind_final_velocity**2 * wind_params.wind_mass_loss_rate / V

    # better encompass the change in rho
    gamma = 5/3
    dp_dt = pressure_from_energy(dE_dt, primitive_state[0, 2:num_injection_cells + 2], primitive_state[1, 2:num_injection_cells + 2], gamma)

    source_term = source_term.at[0, 2:num_injection_cells + 2].set(drho_dt)
    source_term = source_term.at[2, 2:num_injection_cells + 2].set(dp_dt)

    return source_term


import jax.numpy as jnp
from functools import partial
import jax

OPEN_BOUNDARY = 0
REFLECTIVE_BOUNDARY = 1

@partial(jax.jit, static_argnames=['left_boundary', 'right_boundary'])
def _boundary_handler(primitive_states, left_boundary, right_boundary):
    if left_boundary == OPEN_BOUNDARY:
        primitive_states = _open_left_boundary(primitive_states)
    elif left_boundary == REFLECTIVE_BOUNDARY:
        primitive_states = _reflective_left_boundary(primitive_states)
    else:
        raise ValueError("Unknown left boundary condition")
    
    if right_boundary == OPEN_BOUNDARY:
        primitive_states = _open_right_boundary(primitive_states)
    elif right_boundary == REFLECTIVE_BOUNDARY:
        primitive_states = _reflective_right_boundary(primitive_states)
    else:
        raise ValueError("Unknown right boundary condition")
    
    return primitive_states

@jax.jit
def _open_right_boundary(primitive_states):
    primitive_states = primitive_states.at[:, -1].set(primitive_states[:, -3])
    primitive_states = primitive_states.at[:, -2].set(primitive_states[:, -3])
    return primitive_states

@jax.jit
def _open_left_boundary(primitive_states):
    primitive_states = primitive_states.at[:, 0].set(primitive_states[:, 2])
    primitive_states = primitive_states.at[:, 1].set(primitive_states[:, 2])
    return primitive_states

@jax.jit
def _reflective_left_boundary(primitive_states):

    primitive_states = primitive_states.at[:, 0].set(primitive_states[:, 2])
    primitive_states = primitive_states.at[1, 0].set(-primitive_states[1, 2])

    primitive_states = primitive_states.at[:, 1].set(primitive_states[:, 2])
    primitive_states = primitive_states.at[1, 1].set(-primitive_states[1, 2])

    return primitive_states

@jax.jit
def _reflective_right_boundary(primitive_states):

    primitive_states = primitive_states.at[:, -1].set(primitive_states[:, -3])
    primitive_states = primitive_states.at[1, -1].set(-primitive_states[1, -3])

    primitive_states = primitive_states.at[:, -2].set(primitive_states[:, -3])
    primitive_states = primitive_states.at[1, -2].set(-primitive_states[1, -3])

    return primitive_states


import jax.numpy as jnp
import jax
from functools import partial

# The default state jf1uids operates on 
# are the primitive variables rho, u, p.

# ======= Create the primitive state ========

@jax.jit
def primitive_state(rho, u, p):
    return jnp.stack([rho, u, p], axis = 0)

@jax.jit
def density(state):
    return state[0]

@jax.jit
def velocity(state):
    return state[1]

@jax.jit
def pressure(state):
    return state[2]

@jax.jit
def primitive_state_from_conserved(conserved_state, gamma):
    rho, m, E = conserved_state
    u = m / rho
    p = pressure_from_energy(E, rho, u, gamma)
    return jnp.stack([rho, u, p], axis = 0)

# ===========================================

# ======= Create the conserved state ========

@jax.jit
def conserved_state(primitive_state, gamma):
    rho, u, p = primitive_state
    E = total_energy_from_primitives(rho, u, p, gamma)
    return jnp.array([rho, rho * u, E])

# ===========================================

# =============== Fluid physics ===============

@jax.jit
def pressure_from_internal_energy(e, rho, gamma):
  return (gamma - 1) * rho * e

@jax.jit
def internal_energy_from_energy(E, rho, u):
    return E / rho - 0.5 * u**2

@jax.jit
def pressure_from_energy(E, rho, u, gamma):
    e = internal_energy_from_energy(E, rho, u)
    return pressure_from_internal_energy(e, rho, gamma)

@jax.jit
def total_energy_from_primitives(rho, u, p, gamma):
    return p / (gamma - 1) + 0.5 * rho * u**2

@jax.jit
def speed_of_sound(rho, p, gamma):
    return jnp.sqrt(gamma * p / rho)

# ===========================================

# ============= Total Quantities =============

@partial(jax.jit, static_argnames=['num_ghost_cells'])
def calculate_total_mass_proxy(primitive_state, helper_data, dx, num_ghost_cells):
    """
    Calculate the total mass in the domain.
    """
    return jnp.sum(primitive_state[0, num_ghost_cells:-num_ghost_cells] * helper_data.r_hat_alpha[num_ghost_cells:-num_ghost_cells] * dx)

@partial(jax.jit, static_argnames=['num_ghost_cells'])
def calculate_total_energy_proxy(primitive_state, helper_data, dx, gamma, num_ghost_cells):
    """
    Calculate the total energy in the domain.
    """
    energy = total_energy_from_primitives(primitive_state[0, num_ghost_cells:-num_ghost_cells], primitive_state[1, num_ghost_cells:-num_ghost_cells], primitive_state[2, num_ghost_cells:-num_ghost_cells], gamma)
    return jnp.sum(energy * helper_data.r_hat_alpha[num_ghost_cells:-num_ghost_cells] * dx)


import jax.numpy as jnp
import jax

from functools import partial

from jf1uids.geometry.boundaries import _boundary_handler
from jf1uids.fluid_equations.fluid import conserved_state, primitive_state_from_conserved, speed_of_sound
from jf1uids.spatial_reconstruction.limiters import _minmod
from jf1uids.riemann_solver.hll import _hll_solver

@partial(jax.jit, static_argnames=['alpha_geom'])
def calculate_limited_gradients(primitive_states, dx, alpha_geom, rv):
    if alpha_geom == 0:
        cell_distances_left = dx # distances r_i - r_{i-1}
        cell_distances_right = dx # distances r_{i+1} - r_i
    else:
        # calculate the distances
        cell_distances_left = rv[1:-1] - rv[:-2]
        cell_distances_right = rv[2:] - rv[1:-1]

    # formulation 1:
    # a = (primitive_states[:, 1:-1] - primitive_states[:, :-2]) / cell_distances_left
    # b = (primitive_states[:, 2:] - primitive_states[:, 1:-1]) / cell_distances_right
    # g = jnp.where(a != 0, jnp.divide(b, a), jnp.zeros_like(a))
    # slope_limited = jnp.maximum(0, jnp.minimum(1, g)) # minmod
    # limited_gradients = slope_limited * a

    # formulation 2:
    limited_gradients = _minmod(
        (primitive_states[:, 1:-1] - primitive_states[:, :-2]) / cell_distances_left,
        (primitive_states[:, 2:] - primitive_states[:, 1:-1]) / cell_distances_right
    )
    
    return limited_gradients

@partial(jax.jit, static_argnames=['alpha_geom', 'first_order_fallback'])
def reconstruct_at_interface(primitive_states, dt, dx, gamma, alpha_geom, first_order_fallback, helper_data):

    # get fluid variables for convenience
    rho, u, p = primitive_states

    if alpha_geom == 0:
        distances_to_left_interfaces = dx / 2 # distances r_i - r_{i-1/2}
        distances_to_right_interfaces = dx / 2 # distances r_{i+1/2} - r_i
    else:
        r = helper_data.geometric_centers
        rv = helper_data.volumetric_centers

        distances_to_left_interfaces = rv[1:-1] - (r[1:-1] - dx / 2)
        distances_to_right_interfaces = (r[1:-1] + dx / 2) - rv[1:-1]

    # get the limited gradients on the cells
    limited_gradients = calculate_limited_gradients(primitive_states, dx, alpha_geom, helper_data.volumetric_centers)

    # fallback to 1st order
    if first_order_fallback:
        limited_gradients = jnp.zeros_like(limited_gradients)

    # calculate the sound speed
    c = speed_of_sound(rho, p, gamma)

    # calculate the vectors making up A_W
    A_W_1 = jnp.stack([u, jnp.zeros_like(u), jnp.zeros_like(u)], axis = 0)
    A_W_2 = jnp.stack([rho, u, rho * c ** 2], axis = 0)
    A_W_3 = jnp.stack([jnp.zeros_like(u), 1 / rho, u], axis = 0)
    # maybe better construct A_W as a 3x3xN tensor
    # A_W = jnp.stack([A_W_1, A_W_2, A_W_3], axis = 0)

    projected_gradients = A_W_1[:, 1:-1] * limited_gradients[0, :] + A_W_2[:, 1:-1] * limited_gradients[1, :] + A_W_3[:, 1:-1] * limited_gradients[2, :]

    # predictor step
    predictors = primitive_states.at[:, 1:-1].add(-dt / 2 * projected_gradients)

    # compute primitives at the interfaces
    primitives_left = predictors.at[:, 1:-1].add(-distances_to_left_interfaces * limited_gradients)
    primitives_right = predictors.at[:, 1:-1].add(distances_to_right_interfaces * limited_gradients)

    # the first entries are the state to the left and right
    # of the interface between cell 1 and 2
    return primitives_right[:, 1:-2], primitives_left[:, 2:-1]

@partial(jax.jit, static_argnames=['alpha_geom'])
def pressure_nozzling_source(primitive_states, dx, r, rv, r_hat_alpha, alpha_geom):
    _, _, p = primitive_states

    # calculate the limited gradients on the cells
    _, _, dp_dr = calculate_limited_gradients(primitive_states, dx, alpha_geom, rv)

    pressure_nozzling = r[1:-1] ** (alpha_geom - 1) * p[1:-1] + (r_hat_alpha[1:-1] - rv[1:-1] * r[1:-1] ** (alpha_geom - 1)) * dp_dr

    return jnp.stack([jnp.zeros_like(pressure_nozzling), alpha_geom * pressure_nozzling, jnp.zeros_like(pressure_nozzling)], axis = 0)

@partial(jax.jit, static_argnames=['config'])
def evolve_state(primitive_states, dx, dt, gamma, config, helper_data):
    """
    Time derivative of the conserved variables in the MUSCL scheme
    """

    primitive_states = _boundary_handler(primitive_states, config.left_boundary, config.right_boundary)
    
    # initialize the conservative derivative
    conservative_deriv = jnp.zeros_like(primitive_states)

    # get the left and right states at the interfaces
    primitives_left_of_interface, primitives_right_of_interface = reconstruct_at_interface(primitive_states, dt, dx, gamma, config.alpha_geom, config.first_order_fallback, helper_data)

    # calculate the fluxes at the interfaces
    fluxes = _hll_solver(primitives_left_of_interface, primitives_right_of_interface, gamma)

    # update the conserved variables using the fluxes
    if config.alpha_geom == 0:
        conservative_deriv = conservative_deriv.at[:, config.num_ghost_cells:-config.num_ghost_cells].add(-1 / dx * (fluxes[:, 1:] - fluxes[:, :-1]))
    else:
        r = helper_data.geometric_centers
        rv = helper_data.volumetric_centers
        r_hat_alpha = helper_data.r_hat_alpha

        alpha = config.alpha_geom

        r_plus_half = r.at[config.num_ghost_cells:-config.num_ghost_cells].add(dx / 2)
        r_minus_half = r.at[config.num_ghost_cells:-config.num_ghost_cells].add(-dx / 2)

        # calculate the source terms
        nozzling_source = pressure_nozzling_source(primitive_states, dx, r, rv, r_hat_alpha, config.alpha_geom)

        # update the conserved variables using the fluxes and source terms
        conservative_deriv = conservative_deriv.at[:, config.num_ghost_cells:-config.num_ghost_cells].add(1 / r_hat_alpha[config.num_ghost_cells:-config.num_ghost_cells] * (
            - (r_plus_half[config.num_ghost_cells:-config.num_ghost_cells] ** alpha * fluxes[:, 1:] - r_minus_half[config.num_ghost_cells:-config.num_ghost_cells] ** alpha * fluxes[:, :-1]) / dx
            + nozzling_source[:, 1:-1]
        ))

    # update the conservative variables
    conservative_states = conserved_state(primitive_states, gamma)

    # update the conservative variables, here with an Euler step
    conservative_states = conservative_states + dt * conservative_deriv

    # update the primitive variables
    primitive_states = primitive_state_from_conserved(conservative_states, gamma)

    return primitive_states


import jax.numpy as jnp
from jf1uids.fluid_equations.euler import _euler_flux
from jf1uids.fluid_equations.fluid import speed_of_sound
import jax

@jax.jit
def _cfl_time_step(primitive_states, dx, dt_max, gamma, C_CFL = 0.8):
    primitives_left = primitive_states[:, :-1]
    primitives_right = primitive_states[:, 1:]

    rho_L, u_L, p_L = primitives_left
    rho_R, u_R, p_R = primitives_right

    # calculate the sound speeds
    c_L = speed_of_sound(rho_L, p_L, gamma)
    c_R = speed_of_sound(rho_R, p_R, gamma)
    
    # very simple approach for the wave velocities
    wave_speeds_right_plus = jnp.maximum(jnp.maximum(u_L + c_L, u_R + c_R), 0)
    wave_speeds_left_minus = jnp.minimum(jnp.minimum(u_L - c_L, u_R - c_R), 0)

    # calculate the maximum wave speed
    # get the maximum wave speed
    max_wave_speed = jnp.maximum(jnp.max(jnp.abs(wave_speeds_right_plus)), jnp.max(jnp.abs(wave_speeds_left_minus)))

    # calculate the time step
    dt = C_CFL * dx / max_wave_speed

    return jnp.minimum(dt, dt_max)



import jax
from functools import partial

CARTESIAN = 0
CYLINDRICAL = 1
SPHERICAL = 2

@partial(jax.jit, static_argnames=['alpha_geom'])
def _r_hat_alpha(r, dr, alpha_geom):
    if alpha_geom == 2:
        return r ** 2 + 1/12 * dr ** 2
    elif alpha_geom == 1:
        return r
    else:
        raise ValueError("Unknown geometry / not for cartesian coordinates")
    
@partial(jax.jit, static_argnames=['alpha_geom'])
def _center_of_volume(r, dr, alpha_geom):
    if alpha_geom == 1:
        return (r ** 2 + 1/12 * dr ** 2) / r ** 2 * r
    elif alpha_geom == 2:
        r_hat = _r_hat_alpha(r, dr, alpha_geom)
        return (r ** 2 + 1/4 * dr ** 2) / r_hat * r
    


import jax.numpy as jnp
from jf1uids.fluid_equations.fluid import total_energy_from_primitives
import jax

@jax.jit
def _euler_flux(primitive_states, gamma):
    rho, u, p = primitive_states
    m = rho * u
    E = total_energy_from_primitives(rho, u, p, gamma)
    return jnp.stack([m, m * u + p, u * (E + p)], axis=0)




import jax
import jax.numpy as jnp
from functools import partial

from equinox.internal._loop.checkpointed import checkpointed_while_loop

from jf1uids.time_stepping.CFL import _cfl_time_step
from jf1uids.fluid_equations.fluid import calculate_total_energy_proxy, calculate_total_mass_proxy
from jf1uids.spatial_reconstruction.muscl_scheme import evolve_state
from jf1uids.physics_modules.physical_sources import add_physical_sources
from jf1uids.data_classes.simulation_checkpoint_data import CheckpointData

from timeit import default_timer as timer

@partial(jax.jit, static_argnames=['config'])
def time_integration(primitive_state, config, params, helper_data):

    # set dx appropriately
    config = config._replace(dx = config.box_size / (config.num_cells - 1))

    if config.fixed_timestep:
        return time_integration_fixed_steps(primitive_state, config, params, helper_data)
    else:
        if config.adaptive_timesteps_backwards_differentiable:
            return time_integration_adaptive_backwards(primitive_state, config, params, helper_data)
        else:
            return time_integration_adaptive_steps(primitive_state, config, params, helper_data)

@partial(jax.jit, static_argnames=['config'])
def time_integration_fixed_steps(primitive_state, config, params, helper_data):

    if config.intermediate_saves:
        raise NotImplementedError("intermediate_saves not implemented")

    dt = params.t_end / config.num_timesteps

    def update_step(_, state):

        state = add_physical_sources(state, dt / 2, config, params, helper_data)
        state = evolve_state(state, config.dx, dt, params.gamma, config, helper_data)
        state = add_physical_sources(state, dt / 2, config, params, helper_data)

        return state
    
    # use lax fori_loop to unroll the loop
    state = jax.lax.fori_loop(0, config.num_timesteps, update_step, primitive_state)

    return state  


@partial(jax.jit, static_argnames=['config'])
def time_integration_adaptive_steps(primitive_state, config, params, helper_data):

    if config.intermediate_saves:
        times = jnp.zeros(config.num_saves)
        states = jnp.zeros((config.num_saves, primitive_state.shape[0], primitive_state.shape[1]))
        total_mass_proxy = jnp.zeros(config.num_saves)
        total_energy_proxy = jnp.zeros(config.num_saves)
        current_checkpoint = 0
        checkpoint_data = CheckpointData(times = times, states = states, total_mass_proxy = total_mass_proxy, total_energy_proxy = total_energy_proxy, current_checkpoint = current_checkpoint)

    def update_step(carry):

        if config.intermediate_saves:
            time, state, checkpoint_data = carry

            def update_checkpoint_data(checkpoint_data):
                times = checkpoint_data.times.at[checkpoint_data.current_checkpoint].set(time)
                states = checkpoint_data.states.at[checkpoint_data.current_checkpoint].set(state)
                total_mass_proxy = checkpoint_data.total_mass_proxy.at[checkpoint_data.current_checkpoint].set(calculate_total_mass_proxy(state, helper_data, config.dx, config.num_ghost_cells))
                total_energy_proxy = checkpoint_data.total_energy_proxy.at[checkpoint_data.current_checkpoint].set(calculate_total_energy_proxy(state, helper_data, config.dx, params.gamma, config.num_ghost_cells))
                current_checkpoint = checkpoint_data.current_checkpoint + 1
                checkpoint_data = checkpoint_data._replace(times = times, states = states, total_mass_proxy = total_mass_proxy, total_energy_proxy = total_energy_proxy, current_checkpoint = current_checkpoint)
                return checkpoint_data
            
            def dont_update_checkpoint_data(checkpoint_data):
                return checkpoint_data

            checkpoint_data = jax.lax.cond(time >= checkpoint_data.current_checkpoint * params.t_end / config.num_saves, update_checkpoint_data, dont_update_checkpoint_data, checkpoint_data)

            num_iterations = checkpoint_data.num_iterations + 1
            checkpoint_data = checkpoint_data._replace(num_iterations = num_iterations)

        else:
            time, state = carry

        dt = _cfl_time_step(state, config.dx, params.dt_max, params.gamma, params.C_cfl)

        state = add_physical_sources(state, dt / 2, config, params, helper_data)
        state = evolve_state(state, config.dx, dt, params.gamma, config, helper_data)
        state = add_physical_sources(state, dt / 2, config, params, helper_data)

        time += dt

        if config.intermediate_saves:
            carry = (time, state, checkpoint_data)
        else:
            carry = (time, state)

        return carry
    
    def condition(carry):
        if config.intermediate_saves:
            t, _, _ = carry
        else:
            t, _ = carry
        return t < params.t_end
    
    if config.intermediate_saves:
        carry = (0.0, primitive_state, checkpoint_data)
    else:
        carry = (0.0, primitive_state)
    
    start = timer()
    carry = jax.lax.while_loop(condition, update_step, carry)
    end = timer()
    duration = end - start

    if config.intermediate_saves:
        _, state, checkpoint_data = carry
        checkpoint_data = checkpoint_data._replace(runtime = duration)
        return checkpoint_data
    else:
        _, state = carry
        return state
    

@partial(jax.jit, static_argnames=['config'])
def time_integration_adaptive_backwards(primitive_state, config, params, helper_data):

    def update_step(carry):

        time, state = carry

        dt = _cfl_time_step(state, config.dx, params.dt_max, params.gamma, params.C_cfl)

        state = add_physical_sources(state, dt / 2, config, params, helper_data)
        state = evolve_state(state, config.dx, dt, params.gamma, config, helper_data)
        state = add_physical_sources(state, dt / 2, config, params, helper_data)

        time += dt

        carry = (time, state)

        return carry
    
    def condition(carry):
        t, _ = carry
        return t < params.t_end
    
    carry = (0.0, primitive_state)
    
    carry = checkpointed_while_loop(condition, update_step, carry, checkpoints=config.num_checkpoints)

    _, state = carry
    return state


from typing import NamedTuple

# The simulation configuration are parameters defining the simulation
# where changes of necessitate recompilation.

class SimulationConfig(NamedTuple):
    # Simulation parameters
    alpha_geom: int = 0 # 0 -> cartesian, 1 -> cylindrical, 2 -> spherical
    box_size: float = 1.0
    num_cells: int = 400

    # TODO: set num_ghost_cells automatically
    reconstruction_order: int = 1
    # The reconstruction order is the number of 
    # cells on each side of the cell of interest
    # used to calculate the gradients for the
    # reconstruction at the interfaces.

    #                                |---------|
    #                           |---------|
    # stencil              |---------|
    # cells            || 1g | 2g | 3c | 4g | 5g ||
    # reconstructions        |L  R|L  R|L  R|    |
    # fluxes                     -->  -->
    # update                      | 3c'|
    # --> all others are ghost cells

    num_ghost_cells: int = reconstruction_order + 1

    # HAS TO BE UPDATED MANUALLY
    dx: float = box_size / (num_cells - 1)

    left_boundary: int = 0 # 0 -> open, 1 -> reflective
    right_boundary: int = 0 # 0 -> open, 1 -> reflective

    # if you want to use a fixed timestep
    # this changes how the simulation is
    # compiled, mind the CFL criterion
    fixed_timestep: bool = False
    num_timesteps: int = 1000
    adaptive_timesteps_backwards_differentiable: bool = False
    num_checkpoints: int = 100

    stellar_wind: bool = False

    # intermediate saving of the simulation
    intermediate_saves: bool = False
    num_saves: int = 10

    first_order_fallback: bool = False
    # TODO: add more configs


from typing import NamedTuple
import jax.numpy as jnp
from jf1uids.geometry.geometry import _center_of_volume, _r_hat_alpha

# Helper data like the radii and cell volumes 
# in the simulation or cooling tables etc.

class HelperData(NamedTuple):
    geometric_centers: jnp.ndarray = None
    volumetric_centers: jnp.ndarray = None
    r_hat_alpha: jnp.ndarray = None

def get_helper_data(config):
    dx = config.box_size / (config.num_cells - 1)
    if config.alpha_geom == 0:
        r = jnp.linspace(0, config.box_size, config.num_cells)
        r_hat = dx * jnp.ones_like(r) # not really
        return HelperData(geometric_centers = r, r_hat_alpha = r_hat)
    elif config.alpha_geom == 2 or config.alpha_geom == 1:
        # r = jnp.linspace(- 3 * dx/2, config.box_size - 3 * dx / 2, config.num_cells)
        r = jnp.linspace(dx / 2, config.box_size + dx / 2, config.num_cells)
        volumetric_centers = _center_of_volume(r, dx, config.alpha_geom)
        r_hat = _r_hat_alpha(r, dx, config.alpha_geom)
        return HelperData(geometric_centers = r, volumetric_centers = volumetric_centers, r_hat_alpha = r_hat)


import jax.numpy as jnp

def shock_sensor(primitive_state):
    """
    WENO-JS smoothness indicator for shock detection.
    """
    _, _, p = primitive_state
    return (1/4 * (p[2:] - p[:-2])**2 + 13 / 12 * (p[2:] - 2 * p[1:-1] + p[:-2]))

def strongest_shock_radius(primitive_state, helper_data, padL, padR):
    """
    Compute the radius of the strongest shock in the domain.
    """
    r = helper_data.geometric_centers[padL + 1: -padR - 1]
    shock_sensors = shock_sensor(primitive_state)[padL: -padR]
    max_shock_idx = jnp.argmax(shock_sensors)
    return (r[max_shock_idx - 1] * shock_sensors[max_shock_idx - 1] + r[max_shock_idx] * shock_sensors[max_shock_idx] + r[max_shock_idx + 1] * shock_sensors[max_shock_idx + 1]) / (shock_sensors[max_shock_idx - 1] + shock_sensors[max_shock_idx] + shock_sensors[max_shock_idx + 1])

def strongest_shock_dissipated_energy_flux(primitive_state, gamma, padL, padR, statePad):
    """
    Compute the dissipated energy flux due to the strongest shock in the domain.
    """
    shock_sensors = shock_sensor(primitive_state)[padL: -padR]
    primitive_state = primitive_state[:, padL + 1: -padR - 1]
    max_shock_idx = jnp.argmax(shock_sensors)
    
    # get the pre and post shock states
    rho1, v1, P1 = primitive_state[:, max_shock_idx + statePad]
    rho2, v2, P2 = primitive_state[:, max_shock_idx - statePad]

    c1 = jnp.sqrt(gamma * P1 / rho1)

    e_th1 = P1 / ((gamma - 1))
    e_th2 = P2 / ((gamma - 1))

    x_s = rho2 / rho1

    e_diss = e_th2 - e_th1 * x_s ** gamma

    M_1_sq = (P2 / P1 - 1) * x_s / (gamma * (x_s - 1))

    f_diss = e_diss * jnp.sqrt(M_1_sq) * c1 / x_s

    return f_diss
